extern crate base64;
extern crate hex;
mod char_score;

use std::cmp::min;
use std::convert::From;
use std::convert::Into;
use std::iter::FromIterator;
use std::iter::Iterator;
use std::ops::Add;
use std::ops::{Deref, DerefMut};

pub fn hamming_distance<'a, I>(l: I, r: I) -> u32
where
    I: Iterator<Item = &'a u8>,
{
    l.zip(r)
        .fold(0u32, |acc, (x, y)| acc + (*x ^ *y).count_ones())
}

#[derive(Debug, PartialEq, Clone)]
pub struct Bytes {
    pub data: Vec<u8>,
}

impl Bytes {
    pub fn new() -> Bytes {
        Bytes { data: Vec::new() }
    }
    pub fn from_iter<'a, I>(i: I) -> Bytes
    where
        I: Iterator<Item = &'a u8>,
    {
        Bytes {
            data: i.cloned().collect(),
        }
    }
    pub fn from_string(b: &str) -> Self {
        Bytes {
            data: b.bytes().collect(),
        }
    }
    pub fn from_base64(b: &str) -> Self {
        Bytes {
            data: base64::decode(b).unwrap(),
        }
    }
    pub fn from_hex(h: &str) -> Self {
        Bytes {
            data: hex::decode(h).unwrap(),
        }
    }
    pub fn to_base64(&self) -> String {
        base64::encode(&self.data)
    }
    pub fn to_hex(&self) -> String {
        hex::encode(&self.data)
    }
    pub fn to_string(&self) -> Result<String, std::string::FromUtf8Error> {
        String::from_utf8(self.data.clone())
    }
    pub fn xor_fixed(&self, o: &[u8]) -> Result<Bytes, String> {
        if self.data.len() != o.len() {
            return Err(String::from("Length mismatch"));
        }

        return Ok(Bytes {
            data: self
                .data
                .iter()
                .zip(o.iter())
                .map(|x| *x.0 ^ *x.1)
                .collect(),
        });
    }

    pub fn xor_byte(self, o: u8) -> Self {
        Bytes {
            data: self.data.into_iter().map(|x| x ^ o).collect(),
        }
    }

    pub fn xor_repeating(&self, o: &Self) -> Self {
        let len = std::cmp::max(self.len(), o.len());
        Bytes {
            data: self
                .iter()
                .cycle()
                .zip(o.iter().cycle())
                .take(len)
                .map(|(x, y)| *x ^ *y)
                .collect(),
        }
    }

    pub fn score(&self) -> f32 {
        char_score::CharScore::new().score(self.iter())
    }

    pub fn find_best_single_xor_cipher(&self) -> (u8, f32) {
        let mut best_score: f32 = -1000.;
        let mut best_byte: u8 = 0;
        let scorer = char_score::CharScore::new();

        for b in 0..=255 {
            let work_arr = self.clone().xor_byte(b);
            let score = scorer.score(work_arr.iter());
            if score > best_score {
                best_score = score;
                best_byte = b;
            }
        }

        (best_byte, best_score)
    }

    pub fn hamming_distance(&self, o: &Self) -> u32 {
        hamming_distance(self.iter(), o.iter())
    }

    pub fn find_best_xor_cipher(&self) -> Self {
        let mut best_score = (8 * self.len()) as f32;
        let mut best_keysize: usize = 0;
        for keysize in 2..min(45, self.len() / 2) {
            let chunks = self.data.chunks_exact(keysize);
            let left = chunks.clone().step_by(2);
            let right = chunks.skip(1).step_by(2);
            let (num, score) = left.zip(right).fold((0u32, 0u32), |(n, s), (l, r)| {
                (n + 1, s + hamming_distance(l.iter(), r.iter()))
            });
            let score = score as f32 / num as f32 / keysize as f32;
            if score < best_score {
                best_keysize = keysize;
                best_score = score;
            }
        }
        let single_xors = (0..best_keysize).map(|n| self.iter().skip(n).step_by(best_keysize));
        Bytes {
            data: single_xors
                .map(|crypt| Bytes::from_iter(crypt).find_best_single_xor_cipher().0)
                .collect(),
        }
    }

    pub fn pad(mut self, num: u8) -> Self {
        for _ in 0..num {
            self.data.push(num);
        }
        self
    }

    pub fn pad_to_multiple(self, num: usize) -> Self {
        let num = (num - (self.len() % num)) as u8;
        self.pad(num)
    }
}

impl Add for Bytes {
    type Output = Self;

    fn add(self, other: Self) -> Self {
        Self {
            data: Vec::from_iter(self.iter().chain(other.iter()).map(|x| *x)),
        }
    }
}

impl Deref for Bytes {
    type Target = Vec<u8>;

    fn deref(&self) -> &Self::Target {
        &self.data
    }
}

impl DerefMut for Bytes {
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.data
    }
}

impl From<&[u8]> for Bytes {
    fn from(data: &[u8]) -> Self {
        Bytes { data: data.to_vec() }
    }
}

impl From<Vec<u8>> for Bytes {
    fn from(data: Vec<u8>) -> Self {
        Bytes { data }
    }
}

impl Into<Vec<u8>> for Bytes {
    fn into(self) -> Vec<u8> {
        self.data
    }
}
