use crate::bytes::Bytes;

pub use openssl::symm::Mode;

pub struct Aes {
    key: Vec<u8>,
}

impl Aes {
    const BLOCK_SIZE: usize = 16;

    pub fn new(key: &[u8]) -> Option<Aes> {
        let len = key.len();
        if len == 16 || len == 24 || len == 32 {
            return Some(Aes { key: key.to_vec() });
        };
        None
    }

    fn cipher(&self) -> openssl::symm::Cipher {
        match self.key.len() {
            16 => Some(openssl::symm::Cipher::aes_128_ecb()),
            24 => Some(openssl::symm::Cipher::aes_192_ecb()),
            32 => Some(openssl::symm::Cipher::aes_256_ecb()),
            _ => None,
        }
        .unwrap()
    }

    fn crypter(&self, mode: openssl::symm::Mode) -> openssl::symm::Crypter {
        let cipher = self.cipher();
        openssl::symm::Crypter::new(cipher, mode, &self.key, None).unwrap()
    }

    pub fn ecb(&self, mut data: Vec<u8>, mode: Mode) -> Option<Vec<u8>> {
        if data.len() % Self::BLOCK_SIZE != 0 {
            return None;
        }

        let mut crypter = self.crypter(mode);
        let mut output_buffer = [0u8; 2 * Self::BLOCK_SIZE];
        for chunk in data.chunks_exact_mut(Self::BLOCK_SIZE) {
            crypter.update(chunk, &mut output_buffer);
            chunk.copy_from_slice(&output_buffer[0..Self::BLOCK_SIZE]);
        }
        Some(data)
    }

    fn xor(a: &mut [u8], b: &[u8]) {
        for (l, r) in a.iter_mut().zip(b.iter()) {
            *l = *l ^ r;
        }
    }

    fn xor_into(a: &[u8], b: &[u8], c: &mut [u8]) {
        for ((t, l), r) in c.iter_mut().zip(a.iter()).zip(b.iter()) {
            *t = l ^ r;
        }
    }

    pub fn cbc(&self, data: Vec<u8>, iv: &[u8], mode: Mode) -> Option<Vec<u8>> {
        let data = crate::bytes::Bytes::from(data);
        let mut data = data.pad_to_multiple(Self::BLOCK_SIZE);
        let mut crypter = self.crypter(mode);
        let mut iv_ = [0u8; Self::BLOCK_SIZE];
        let mut output_buffer = [0u8; 2 * Self::BLOCK_SIZE];

        iv_.clone_from_slice(iv);
        let mut iv = iv_;
        for chunk in data.chunks_exact_mut(Self::BLOCK_SIZE) {
            match mode {
                Mode::Decrypt => {
                    let mut old_chunk = [0u8; Self::BLOCK_SIZE];
                    old_chunk.copy_from_slice(chunk);

                    let len = crypter.update(chunk, &mut output_buffer);
                    Self::xor_into(&output_buffer, &iv, chunk);

                    println!("IV: {:?}\nIB: {:?}\nOB: {:?} ({:?})\n\n", iv, old_chunk, chunk, len);

                    iv.clone_from_slice(&old_chunk);
                }
                Mode::Encrypt => {
                    Self::xor(chunk, &iv);

                    crypter.update(chunk, &mut output_buffer);
                    chunk.copy_from_slice(&output_buffer[0..Self::BLOCK_SIZE]);

                    iv.copy_from_slice(chunk);
                }
            }
        }
        println!("{:?}", data);
        Some(data.data.clone())
    }
}
